import pytest
import dask.dataframe as dd
import dask_quik as dq


def is_dc_dd(dc_ddf, dc_ddt):
    if bool(dc_ddt):
        import dask_cudf

        assert isinstance(dc_ddf, dask_cudf.DataFrame)
    else:
        assert isinstance(dc_ddf, dd.DataFrame)


def test_scatter_gpu(sample_data, args):
    """create a dc_dd using scatter and gpu"""
    if bool(args.gpus) and not args.has_gpu:
        with pytest.raises(ImportError):
            import cudf
        return
    dc_ddf = dq.transform.scatter_and_gpu(sample_data, args)
    is_dc_dd(dc_ddf, args.gpus)


def test_gpu_sort_cpu(sample_data, args):
    """create a dc_dd using scatter_and_gpu.
    If gpus=1, scatter_and_gpu will create a
    dask_cudf.DataFrame, then it will be sorted
    and sent back to the CPU. If gpus=0, we'll
    do the same but all on CPU."""
    df = sample_data.copy()
    if bool(args.gpus) and not args.has_gpu:
        with pytest.raises(ImportError):
            import cudf
        return
    dc_ddf = dq.transform.scatter_and_gpu(df, args)
    is_dc_dd(dc_ddf, args.gpus)
    df = df.set_index("item_id").sort_index()
    if bool(args.gpus):
        ddf = dq.transform.gpu_sort_cpu(dc_ddf, "item_id")
        print(ddf.compute())
    else:
        ddf = dc_ddf.set_index("item_id")
        print(ddf.compute())
    assert ddf.compute().equals(df)


def test_dc_sort_index(sample_data, args):
    """create a dc_dd using scatter and gpu"""
    if bool(args.gpus) and not args.has_gpu:
        with pytest.raises(ImportError):
            import cudf
        return
    df = sample_data.copy()
    df.index.names = ["ui_index"]
    dc_ddf = dq.transform.scatter_and_gpu(df, args)
    is_dc_dd(dc_ddf, args.gpus)
    if bool(args.gpus):
        dc_ddf = dc_ddf.sort_values("item_id")
        dc_ddf = dq.transform.dc_sort_index(dc_ddf)
        dc_ddf = dc_ddf.compute().to_pandas()
    else:
        dc_ddf = dc_ddf.reset_index().set_index("item_id")
        dc_ddf = dc_ddf.reset_index().set_index("ui_index")
        dc_ddf = dc_ddf[["user_number", "item_id"]].compute()
    assert dc_ddf.equals(df.sort_index())



import pandas as pd
import dask
import dask.dataframe as dd
import dask_ml
from typing import Union
from argparse import Namespace
import warnings
import time
import dask_quik.utils as du
from argparse import Namespace
from typing import Dict, Any

try:
    import dask_cudf as dc
    import cudf
except ImportError:
    warnings.warn(
        "dask_quik.transform unable to import GPU libraries, \
        importing a dummy dc.DataFrame"
    )
    import dask_quik.dummy as dc

# allow the ability of a dask_cudf.DataFrame or a dd.DataFrame
dc_dd = Union[dc.DataFrame, dd.DataFrame]


def tt_leave_one_out(
    ddf: dd.DataFrame, cols: Dict[str, Any], args: Namespace
) -> Dict[str, pd.DataFrame]:
    """Using the leave one out method, "leaving out" the most recent
    record for test, and creating a training set with all the remaining
    records. This is common for recommender systems.

    Args:
        ddf (dd.DataFrame): The original dataframe
        cols (Dict[str, Any]): A dictionary of column keys and values
        args (Namespace): argumnets for train, (valid), test, and seed.

    Returns:
        Dict[str, pd.DataFrame]: A dict of pandas dataframes for train, (valid),
        and test.
    """
    te_ddf = ddf[(ddf.latest == 1) | (ddf.rating == -1)]
    tr_ddf = ddf[~((ddf.latest == 1) | (ddf.rating == -1))]
    tr_ddf, vl_ddf = dask_ml.model_selection.train_test_split(
        tr_ddf,
        train_size=args.train_size,
        test_size=args.test_size,
        random_state=args.seed,
    )
    tkeys = ["train", "test"]
    ddf_list = [tr_ddf, te_ddf]
    if args.valid_size > 0:
        tkeys.append('valid')
        ddf_list.append(vl_ddf)
    colk = ["user", "item", "late", "rate"]
    t_cols = list(du.subdict(cols, colk).values())
    ddf_list = dask.compute(*[ddf[t_cols] for ddf in ddf_list])
    return dict(zip(tkeys, ddf_list))



import os
import time
import dask.dataframe as dd
from dask.distributed import Client
from subprocess import check_output, STDOUT
from argparse import Namespace
from typing import Union, Optional, Dict, Any
import warnings
import dask_quik.utils as du

try:
    import dask_cudf as dc
    import cudf
    from dask_cuda import LocalCUDACluster
except ImportError:
    warnings.warn(
        "dask_quik.utils unable to import GPU libraries, \
        using dask instead of dask_cudf"
    )
    import dask_quik.dummy as dc

dc_dd = Union[dc.DataFrame, dd.DataFrame]


def setup_cluster(use_gpus: Optional[int] = None, worker_space: Optional[str] = None) -> Client:
    """Setup a dask distributed cuda cluster on GPUs. Sometimes
    if this has been done before, the legacy data needs to be
    removed.

    Args:
        use_gpus (int, optional): Number of gpus to use
        worker_space (str, optional): The location of the
        dask-worker-space director. Defaults to None.

    Returns:
        Client: A distributed cluster to contain dask or dask_cudf objects.
    """
    if worker_space is not None and os.path.exists(worker_space):
        os.system(f"rm -r {worker_space}/*")
    if use_gpus is None:
        use_gpus = gpus()
    if bool(use_gpus):
        cluster = LocalCUDACluster(n_workers=use_gpus,
                                   local_directory=worker_space)
        print("GPU cluster has been established", flush=True)
    else:
        cluster = None
        print("No GPUs found - running cluster on CPU.", flush=True)
    return Client(address=cluster)


def subdict(cols: Dict[str, Any], subkeys: list) -> Dict[str, Any]:
    """Take a dictionary and subset it based on a list of keys.

    Args:
        full_dict (Dict[str, Any]): The full dictionary to be subsetted
        subkeys (list): list of keys to be contained in the subset

    Returns:
        Dict[str, Any]: A subsetted dictionary
    """
    return {key: cols[key] for key in subkeys}


def sec_str(start_time: float) -> str:
    """[summary]

    Args:
        start_time (float): output from time.time() earlier in the program

    Returns:
        str: A formatted string with the seconds that a code segment took to run
    """
    return str(round(time.time() - start_time, 2)) + " seconds"


def row_str(dflen: int) -> str:
    """[summary]

    Args:
        dflen (int): [description]

    Returns:
        str: A formatted string with the number of rows (in millions).
    """
    return str(round(dflen / 1000000, 1)) + "M rows"


def shrink_dtypes(df: dc_dd, df_dtypes: dict) -> dc_dd:
    """reduces the data type of a set of series.

    Args:
        df (dc_dd): data frame with series to reduce dtypes
        df_dtypes (dict): key, value pairs of column names and their new data types

    Returns:
        dc_dd: dask_cudf or dd with reduced dtypes
    """
    for colname, newdtype in df_dtypes.items():
        df[colname] = df[colname].astype(newdtype)
    return df


def gpus() -> int:
    """Determines if there are GPUs on the system, and
    how many

    Returns:
        int: number of GPUs on the system
    """
    gpu_cmd = "nvidia-smi -L | wc -l"
    gpus = check_output(gpu_cmd, stderr=STDOUT, shell=True)
    gpus = int(gpus.splitlines()[-1])
    return gpus



import pandas as pd
import dask.dataframe as dd
from typing import Union
from argparse import Namespace
import warnings
import time
import dask_quik.utils as du

try:
    import dask_cudf as dc
    import cudf
except ImportError:
    warnings.warn(
        "dask_quik.transform unable to import GPU libraries, \
        importing a dummy dc.DataFrame"
    )
    import dask_quik.dummy as dc

# allow the ability of a dask_cudf.DataFrame or a dd.DataFrame
dc_dd = Union[dc.DataFrame, dd.DataFrame]


def scatter_and_gpu(cdf: pd.DataFrame, args: Namespace) -> dc_dd:
    """First convert a pandas dataframe to a dask dataframe (scatter)
    and then if there's a gpu, move it to the gpu.

    Args:
        cdf (pd.DataFrame): initial df
        args (Namespace): arguments

    Returns:
        cdf dc_dd: the final df
    """
    cdf = dd.from_pandas(cdf, npartitions=args.partitions)
    if bool(args.gpus):
        cdf = cdf.map_partitions(cudf.DataFrame.from_pandas)
    return cdf


def gpu_sort_cpu(gdf: dc.DataFrame, idx_col: str) -> dd.DataFrame:
    """First sort the dask_cudf DataFrame by an index column,
    then push the partitions to the CPU as a dask DataFrame

    Args:
        gdf (dask_cudf.DataFrame): The unsorted dask_cudf DataFrame on GPU
        idx_col (str): The index column to be sorted

    Returns:
        dd.DataFrame: The sorted, partitioned dask DataFrame on CPU
    """
    st = time.time()
    gdf = gdf.set_index(idx_col)
    cdf = gdf.map_partitions(lambda df: df.to_pandas())
    print("sorted and moved to cpu in " + du.sec_str(st))
    return cdf


def dc_sort_index(dcdf: dc.DataFrame) -> dc.DataFrame:
    """emulate pandas sort_index for a dask_cudf DataFrame.
    This pulls the index into a column, then uses set_index.

    Args:
        dcdf (dc.DataFrame): original dask_cudf DataFrame

    Returns:
        dc.DataFrame: newly sorted index for dask_cudf DataFrame
    """
    idx = dcdf.index.name
    dcdf = dcdf.reset_index().rename(columns={"index": idx})
    dcdf = dcdf.set_index(idx)
    return dcdf



import pytest
import pandas as pd
import dask.dataframe as dd
import dask_quik as dq
import dask_quik.dummy as dc


def t_dc_dd(dc_ddf, dc_ddt):
    if bool(dc_ddt):
        import dask_cudf

        assert isinstance(dc_ddf, dask_cudf.DataFrame)
    else:
        assert isinstance(dc_ddf, dd.DataFrame)


def test_dummy_df(args):
    dum_df = dc.DataFrame()
    # dum_call = dum_df.__call__
    # dum_attr = dum_df.__getattr__


def test_dask_df(sample_data, args):
    """create a dask df"""
    ddf = dd.from_pandas(sample_data, npartitions=args.partitions)
    assert isinstance(ddf, dd.DataFrame)


def test_cudf_df(sample_data, args):
    """create a cudf df"""
    if bool(args.gpus) and not args.has_gpu:
        with pytest.raises(ImportError):
            import cudf
        return
    elif bool(args.gpus):
        import cudf

        cdf = cudf.from_pandas(sample_data)
        assert isinstance(cdf, cudf.DataFrame)
    else:
        dum_df = dc.DataFrame()
        assert isinstance(dum_df, dc.DataFrame)


def test_dask_cudf(sample_data, args):
    """create a dask_cudf df"""
    if bool(args.gpus) and not args.has_gpu:
        with pytest.raises(ImportError):
            import dask_cudf
        return
    elif bool(args.gpus):
        import cudf
        import dask_cudf

        cdf = cudf.from_pandas(sample_data)
        dcdf = dask_cudf.from_cudf(cdf, npartitions=args.partitions)
        assert isinstance(dcdf, dask_cudf.DataFrame)
    else:
        dum_df = dc.DataFrame()
        assert isinstance(dum_df, dc.DataFrame)



import pandas as pd
import numpy as np
import dask.dataframe as dd
import dask.array as da
from typing import Union, Tuple, Optional, Dict, Any, List
from argparse import Namespace
import warnings
import dask_quik.utils as du
import dask_quik.transform as dt
import time

try:
    import dask_cudf as dc
    import cudf
except ImportError:
    warnings.warn(
        "dask_quik.cartesian unable to import GPU libraries, \
        importing a dummy dc.DataFrame"
    )
    import dask_quik.dummy as dc

# allow the ability of a dask_cudf.DataFrame or a dd.DataFrame
dc_dd = Union[dc.DataFrame, dd.DataFrame]


def add_cartesian_dummy(odf: dc_dd) -> dc_dd:
    """create a dummy dataframe with a column "cartesian" with the value
    0 to force a cartesian join using merge.

    Args:
        odf (dc_dd): original dask_cudf or dd

    Returns:
        dc_dd: dummy dask_cudf or dd
    """
    odf["cartesian"] = 0
    odf = du.shrink_dtypes(odf, {"cartesian": "int8"})
    return odf


def dask_cudf_cartesian(odf: dc_dd, colv: List[str], args: Namespace) -> dc_dd:
    """Create an cartesian product of two columns using a dummy inner join.
    This was the best option for dask and cudf. The function:
    - adds a dummy column with value 0
    - creates a distinct list of each column (with dummy)
    - does an inner join of the two lists on the dummy column

    Args:
        odf (dc_dd): dask_cudf or dd to cross-join
        colv (Dict[str, Any]): dictionary of column type and column name
        p (int): number of partitions

    Returns:
        dc_dd: dask_cudf or dd cartesian product of the original dataframe
    """
    odf = add_cartesian_dummy(odf)
    if bool(args.gpus):
        gv_list = [
            odf[[col, "cartesian"]]
            .sort_values(col)
            .drop_duplicates(split_out=args.partitions)
            .reset_index(drop=True)
            for col in colv
        ]
    else:
        gv_list = [
            odf[[col, "cartesian"]]
            .set_index(col)
            .persist()
            .reset_index()
            .drop_duplicates(split_out=args.partitions)
            .reset_index(drop=True)
            for col in colv
        ]
    gv_list[1] = gv_list[1].compute()
    odf = gv_list[0].merge(gv_list[1], how="inner", on="cartesian")
    odf = odf.drop("cartesian", axis=1)
    return odf


def indexize(
    dc: dc_dd, cnts: Dict[str, int], colv: List[str], rmcols: Optional[bool] = False
) -> dc_dd:
    """create an index for the permutation of all values. Usually two columns,
    this takes the max of each, and indexes from 0 to max all the possible permutations.

    Args:
        dc (dc_dd): the dask_cudf or dask df to be indexized
        cnts (Dict[str, int]): the counts of each column for permutations
        colv (List[str]): the columns to be indexized
        rmcols (Optional[bool], optional): whether the original columns should be
        removed. Defaults to False.

    Returns:
        dc_dd: The final indexized dask_cudf or dask dataframe
    """
    col2max = cnts[colv[1]]
    idx = colv[0][:1] + colv[1][:1] + "_index"
    dc[idx] = (dc[colv[0]] - 1) * col2max + dc[colv[1]] - 1
    dc = dc.persist()
    dc = dc.set_index(idx)
    if rmcols:
        dc = dc.drop(colv, axis=1)
    return dc


def sparse_cudf_matrix(
    gdf: dc_dd,
    cols: Dict[str, Any],
    cnts: Dict[str, int],
    colk: List[str],
    args: Namespace,
) -> dc_dd:
    """Create an sparse matrix of true customer data within the total
    universe of customer data (cartesian product).
    - determines column names
    - creates the true set
    - creates the sparse matrix (cudf_cartesian)
    - notes which in the sparse matrix are true, false

    Args:
        gdf (dc_dd): dataframe to create sparse matrix
        cols (Dict[str, Any]): dictionary of column type and column name
        cnts (Dict[str, int]): [description]
        colk (List[str]): [description]
        args (Namespace): [description]

    Returns:
        dc_dd: dask_cudf or dd final sparse matrix
    """
    # column name stuff
    st = time.time()
    tcol = "_".join(colk)
    colv = [cols[k] for k in colk]
    # create the cartesian product
    gdf = gdf[colv].drop_duplicates(split_out=args.partitions)
    sm = dask_cudf_cartesian(gdf.copy(), colv, args)
    print("created the " + tcol + " cartesian product in " + du.sec_str(st), flush=True)
    sm = indexize(sm, cnts, colv)
    # idx = sm.index.name
    gdf = indexize(gdf, cnts, colv)
    gdf[tcol] = True
    if not bool(args.gpus):
        colv = None
        gdf = gdf[[tcol]]
    sm = sm.merge(
        gdf,
        how="left",
        on=colv,
        npartitions=args.partitions,
        left_index=True,
        right_index=True,
    )
    print("indexed and merged in " + du.sec_str(st), flush=True)
    sm = dt.dc_sort_index(sm)
    del gdf
    sm[tcol] = sm[tcol].fillna(False)
    return sm



from pathlib import Path
import pytest
import pandas as pd
import json
from argparse import Namespace
from os import system
import warnings
import dask_quik as dq

TESTDIR = Path(__file__).parent
SAMPLE = TESTDIR.joinpath("sample_data.json")
PROD = TESTDIR.joinpath("prod_data.json")
FINAL = TESTDIR.joinpath("final_data.json")
REAL = TESTDIR.joinpath("real_data.json")


def pytest_addoption(parser):
    # parser.addoption("--gpus", type=int, default=0,
    # help="number of gpus per node")
    parser.addoption(
        "--has_gpu",
        action="store_true",
        default=bool(dq.utils.gpus()),
        dest="has_cpu",
        help="testing on a node with a GPU",
    )


def pytest_generate_tests(metafunc):
    metafunc.parametrize("gpus", [0, 1])


@pytest.fixture
def has_gpu(request):
    """argument whether there actually is a gpu"""
    return request.config.getoption("--has_gpu")


@pytest.fixture
def args(gpus):
    """sample args namespace"""
    args = Namespace()
    args.gpus = gpus
    args.has_gpu = system("nvidia-smi -L") == 0
    if not args.has_gpu:
        warnings.warn(
            "GPU not found, setting has_gpu to False. \
            Some tests will be skipped"
        )
    args.partitions = 2
    return args


@pytest.fixture(scope="module")
def cols_dict():
    """sample column names"""
    cols = {
        "user": "user_number",
        "item": "item_id",
        "prod": "product_id",
        "recent": "recency",
        "late": "latest",
    }
    return cols


@pytest.fixture(scope="module")
def ui_cols(cols_dict):
    """sample column names"""
    uicols = dq.utils.subdict(cols_dict, ["user", "item"])
    return uicols


@pytest.fixture(scope="session")
def sample_data():
    """sample user/item dataset"""
    with open(SAMPLE) as f:
        df = pd.DataFrame(json.load(f))
    return df


@pytest.fixture(scope="session")
def prod_data():
    """sample user/item dataset"""
    with open(PROD) as f:
        df = pd.DataFrame(json.load(f))
    return df


@pytest.fixture(scope="session")
def real_data():
    """sample user/item dataset"""
    with open(REAL) as f:
        df = pd.DataFrame(json.load(f))
    return df


@pytest.fixture(scope="session")
def final_data():
    """final user/item dataset"""
    with open(FINAL) as f:
        df = pd.DataFrame(json.load(f))
    df.index.names = ["ui_index"]
    return df



class DataFrame:
    """A dummy DataFrame object to satisfy typing, and allow for
    dask_cudf code be written alongside dask code.
    """

    def __init__(self, *args, **kwargs):
        pass

    # def __call__(self, *args, **kwargs):
    #     return self

    # def __getattr__(self, *args, **kwargs):
    #     return self



import pytest
import dask_quik as dq
import dask.dataframe as dd
import pandas as pd


def is_dc_dd(dc_ddf, dc_ddt):
    if bool(dc_ddt):
        import dask_cudf

        assert isinstance(dc_ddf, dask_cudf.DataFrame)
    else:
        assert isinstance(dc_ddf, dd.DataFrame)


def test_merge(sample_data, prod_data, cols_dict, args):
    """create two dask or dask_cudf DataFrames
    and then combine them"""
    if bool(args.gpus) and not args.has_gpu:
        with pytest.raises(ImportError):
            import cudf
        return
    smp_df = dq.transform.scatter_and_gpu(sample_data, args)
    is_dc_dd(smp_df, args.gpus)
    prd_df = dq.transform.scatter_and_gpu(prod_data, args)
    is_dc_dd(prd_df, args.gpus)
    sm = dq.combine.left_merge_cols(smp_df, prd_df, cols_dict, ["item"])
    is_dc_dd(sm, args.gpus)
    print(sm)


def test_prune(real_data, cols_dict, args):
    dc = dq.transform.scatter_and_gpu(real_data, args)
    is_dc_dd(dc, dq.utils.gpus())
    dcp = dq.combine.prune_rows(
        dc,
        cols_dict,
        grby_cols=["user", "item", "prod"],
        p=args.partitions,
        max=["recent"],
        min=["late"],
    )
    is_dc_dd(dcp, dq.utils.gpus())
    slist = [45671158, 42599686, 85, -6, 443]
    assert list(dcp.sum().compute()) == slist
    rlist = [411070, 6038, 1, 0, 45]
    testrow = dcp[(dcp.user_number == 411070) & (dcp.product_id == 1)]
    assert list(testrow.compute().T.iloc[:, 0]) == rlist



import pandas as pd
import numpy as np
import dask.dataframe as dd
import dask.array as da
from typing import Union, Tuple, Optional, Dict, Any, List
from argparse import Namespace
import warnings
import dask_quik.utils as du

try:
    import dask_cudf as dc
    import cudf
except ImportError:
    warnings.warn(
        "dask_quik.cartesian unable to import GPU libraries, \
        importing a dummy dc.DataFrame"
    )
    import dask_quik.dummy as dc

# allow the ability of a dask_cudf.DataFrame or a dd.DataFrame
dc_dd = Union[dc.DataFrame, dd.DataFrame]


def left_merge_cols(
    left: dc_dd, right: dc_dd, cols: Dict[str, Any], col_list: List[str]
) -> dc_dd:
    """Simple left merge.

    Args:
        left (dc_dd): left DataFrame to merge
        right (dc_dd): right DataFrame to merge
        cols (Dict[str, Any]): dictionary of all column types and column names
        col_list (List[str]): names of columns to merge on

    Returns:
        dc_dd: final merged dataframe
    """
    on_cols = list(du.subdict(cols, col_list).values())
    left = left.merge(right, how="left", on=on_cols)
    return left


def prune_rows(
    dc: dc_dd,
    cols: Dict[str, Any],
    grby_cols: List[str],
    p: int,
    **agg_groups: Optional[List[str]]
) -> dc_dd:
    """When it's necessary to remove slight duplicates, this
    function will group by certain columms, then take the max of some,
    the min of some, and the average of others.

    Args:
        dc (dc_dd): original dask or dask_cudf DataFrame
        cols (Dict[str, Any]): a dictionary of column keys and values
        grby_cols (List[str]): column keys to be grouped by
        p (int): [description]
        **agg_groups: column key value pairs, where the key is the agg
        type (avg, min, max) and the value is the column key to be
        aggregated

    Returns:
        dc_dd: A final pruned dask or dask_cudf DataFrame
    """
    gr = list(du.subdict(cols, grby_cols).values())
    # First pull the column values, then spread them across the aggs, then
    # combine them into a flat dictionary
    aggs = {
        k: list(du.subdict(cols, v).values()) for k, v in agg_groups.items()
    }
    aggs = [dict.fromkeys(v, k) for k, v in aggs.items()]
    aggs = {k: ad[k] for ad in aggs for k in ad}
    dc = dc.groupby(gr).agg(aggs)[list(aggs.keys())].reset_index()
    return dc


def pr(dc, cols, grby_cols, p: int, **kwargs):
    print(kwargs)



import dask_quik.cartesian
import dask_quik.combine
import dask_quik.transform
import dask_quik.utils
import dask_quik.dummy
import dask_quik.split

__all__ = ["cartesian", "combine", "transform", "dummy", "utils", "split"]
__version__ = "0.0.4"



#!/usr/bin/env python

from setuptools import setup, find_packages
from os import system
from pathlib import Path

here = Path(__file__).parent
README = here.joinpath("README.md").read_text()

with open(here.joinpath("requirements.txt")) as f:
    all_reqs = f.read().split("\n")

removes = ["git+"]
if system("nvidia-smi -L") != 0:
    removes.extend(["cudf", "cuda"])
install_requires = [
    x.strip() for x in all_reqs if not any(y in x for y in removes)
]


setup(
    name="dask-quik",
    version="0.0.4",
    description="function to make working in dask_cudf and dask quik-er",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/donchesworth/dask-quik",
    author="Don Chesworth",
    author_email="donald.chesworth@gmail.com",
    license="BSD",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.8",
    ],
    packages=find_packages(exclude=["tests"]),
    install_requires=install_requires,
)



import pytest
import pandas as pd
import dask.dataframe as dd
import dask_quik as dq


@pytest.fixture(scope="module")
def colk(ui_cols):
    """column keys"""
    colk = list(ui_cols.keys())
    return colk


@pytest.fixture(scope="module")
def colv(ui_cols):
    """column values"""
    colv = list(ui_cols.values())
    return colv


@pytest.fixture(scope="module")
def tcol(ui_cols):
    """two column value"""
    tcol = "_".join(ui_cols.keys())
    return tcol


@pytest.fixture(scope="module")
def counts_dict(sample_data):
    """unique counts for sample_data"""
    counts = {
        "user_number": len(sample_data.iloc[:, 0].unique()),
        "item_id": len(sample_data.iloc[:, 1].unique()),
    }
    return counts


def is_dc_dd(dc_ddf, dc_ddt):
    if bool(dc_ddt):
        import dask_cudf

        assert isinstance(dc_ddf, dask_cudf.DataFrame)
    else:
        assert isinstance(dc_ddf, dd.DataFrame)


def test_cartesian_dc_dd(sample_data, colv, args):
    """create a dask cartesian df. If gpus,
    output should be a dask_cudf df, else dask df"""
    if bool(args.gpus) and not args.has_gpu:
        with pytest.raises(ImportError):
            import cudf
        return
    dc_ddf = dq.transform.scatter_and_gpu(sample_data, args)
    is_dc_dd(dc_ddf, args.gpus)
    sm = dq.cartesian.dask_cudf_cartesian(dc_ddf, colv, args)
    is_dc_dd(dc_ddf, args.gpus)


def test_indexized_ddf(sample_data, counts_dict, colv, args):
    """create an indexed df. If gpus,
    output should be a dask_cudf df, else dask df"""
    ddf = dd.from_pandas(sample_data, npartitions=args.partitions)
    ddf = dq.cartesian.indexize(ddf, counts_dict, colv)
    assert isinstance(ddf, dd.DataFrame)


def test_indexized_dc_dd(sample_data, counts_dict, colv, args):
    """create an indexed df. If gpus,
    output should be a dask_cudf df, else dask df"""
    if bool(args.gpus) and not args.has_gpu:
        with pytest.raises(ImportError):
            import cudf
        return
    dc_ddf = dq.transform.scatter_and_gpu(sample_data, args)
    dc_ddf = dq.cartesian.indexize(dc_ddf, counts_dict, colv)
    is_dc_dd(dc_ddf, args.gpus)


def test_indexized_cartesian(sample_data, counts_dict, colv, args):
    """create an indexed cartesian df. If gpus,
    output should be a dask_cudf df, else dask df"""
    if bool(args.gpus) and not args.has_gpu:
        print("unable to test dask_cudf portion")
        pytest.skip()
    else:
        dc_ddf = dq.transform.scatter_and_gpu(sample_data, args)
        sm = dq.cartesian.dask_cudf_cartesian(dc_ddf, colv, args)
        sm = dq.cartesian.indexize(sm, counts_dict, colv)


def test_sparse_matrix(
    final_data, sample_data, cols_dict, counts_dict, colk, args
):
    """create an indexed cartesian df. If gpus,
    output should be a dask_cudf df, else dask df"""
    if bool(args.gpus) and not args.has_gpu:
        with pytest.raises(ImportError):
            import cudf
        return
    dc_ddf = dq.transform.scatter_and_gpu(sample_data, args)
    sm = dq.cartesian.sparse_cudf_matrix(
        dc_ddf, cols_dict, counts_dict, colk, args
    )
    is_dc_dd(sm, args.gpus)
    if bool(args.gpus):
        sm = dq.transform.dc_sort_index(sm).compute().to_pandas()
    else:
        sm = sm.compute()
    assert sm.equals(final_data)



import pytest
import dask_quik as dq
from pathlib import Path
from time import time, sleep
from pynvml.nvml import NVMLError


def test_setup_cluster(args):
    """setup a dask cluster"""
    print(args)
    worker_path = Path.cwd().joinpath("dask-worker-space")
    worker_path.mkdir(parents=True, exist_ok=True)
    worker_path.joinpath("tmp").touch(exist_ok=True)
    if bool(args.gpus) and not args.has_gpu:
        with pytest.raises(
            (KeyError, OSError, NVMLError, AttributeError, NameError)
        ):
            client = dq.utils.setup_cluster(worker_space=worker_path)
        return
    elif bool(args.gpus):
        client = dq.utils.setup_cluster(worker_space=worker_path)
    else:
        pytest.skip()


def test_subdict(cols_dict, args):
    """create a subdict"""
    tsd = dq.utils.subdict(cols_dict, ["item"])
    fd = {"item": "item_id"}
    assert tsd == fd


def test_sec_str(args):
    """print a sec_str"""
    start_time = time()
    sleep(2)
    print(dq.utils.sec_str(start_time))


def test_row_str(sample_data, args):
    """print a row_str"""
    print(dq.utils.row_str(sample_data.shape[0]))
